/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007, 2008, 2009 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * Implementation of the low-level prime field multiplication functions.
 *
 * @version $Id: relic_fp_mul_low.c 161 2009-11-07 17:18:26Z dfaranha $
 * @ingroup bn
 */

#include "relic_fp.h"
#include "relic_fp_low.h"

#define __MULTIPLICATION_IN_ASM__	   1
//#define __FP_MUL__	   COMBA

/*============================================================================*/
/* Private definitions                                                         */
/*============================================================================*/
void fp_muln_low_asm_comba_446(dig_t *, dig_t *, dig_t *);


/*============================================================================*/
/* Public definitions                                                         */
/*============================================================================*/

dig_t fp_muladd_low(dig_t *c, dig_t *a, dig_t digit) {
	return mpn_addmul_1(c, a, FP_DIGS, digit);
}

void fp_muln_low(dig_t *c, dig_t *a, dig_t *b) {
#if __MULTIPLICATION_IN_ASM__ == 0
	mpn_mul_n(c, a, b, FP_DIGS);
#elif FP_MUL == COMBA
	fp_muln_low_asm_comba_446(c, a, b); //i think this is the one used df
#endif
}

/*
void fp_muln_low_basic(dig_t *c, dig_t *a, dig_t *b) {
#if FP_PRIME == 256 
	fp_muln_low_asm_basic_256(c, a, b);
#else
//	fp_muln_low_asm_basic_gen(c, a, b);
	fp_muln_low_asm_basic_256(c, a, b); // todo
#endif	
}

void fp_muln_low_comba(dig_t *c, dig_t *a, dig_t *b) {
	fp_muln_low_basic(c, a, b); // todo
}
*/

//key field multiplication function ?
void fp_mulm_low(dig_t *c, dig_t *a, dig_t *b) {
	dig_t align t[2 * FP_DIGS];

	fp_muln_low(t, a, b);
	fp_rdc(c, t);
}
